
# Autogenerated by Marvin's setup.py on 2017-08-02 14:25:20.450603
import datetime
from pkg_resources import get_distribution, DistributionNotFound
major = 2
minor = 1
bugfix = 4
release = True
timestamp = datetime.datetime(2017, 8, 2, 14, 25, 20, 450603)
version = '2.1.4'


def get_version():
    try:
        dist = get_distribution('sdss-marvin')
    except DistributionNotFound:
        return version
    else:
        return dist.version
    